""" File Submission Interfaces.

TODO update this

The Submission service encapsulates the core functionality of accepting,
triaging and forwarding a submission to the dispatcher.

There are three primary modes of submission:

  two-phase (presubmit + submit)
  inline  (submit file)
  existing (submit a file that is already in the file cache/SAN)

In two-phase mode, submission is a presubmit followed by a submit.
A 'presubmit' is sent to the submission service first. If the server already
has a copy of the sample it indicates as such to the client which saves the
client from copying the file again. Once the client has copied the file
(if required) it then issues a final 'submit'.

"""
from assemblyline.common.classification import InvalidClassification
import elasticapm
import logging
import os
from typing import List, Tuple, Dict

from assemblyline.common import forge
from assemblyline.common.codec import decode_file
from assemblyline.common.dict_utils import flatten
from assemblyline.common.isotime import epoch_to_iso, now, now_as_iso
from assemblyline.common.str_utils import safe_str
from assemblyline.datastore.helper import AssemblylineDatastore
from assemblyline.filestore import FileStore
from assemblyline.odm.messages.submission import Submission as SubmissionObject
from assemblyline.odm.models.file import File as FileInfo
from assemblyline.odm.models.result import Result
from assemblyline.odm.models.submission import File, Submission
from assemblyline_core.dispatching.client import DispatchClient

Classification = forge.get_classification()


def assert_valid_sha256(sha256):
    if len(sha256) != 64:
        raise ValueError('Invalid SHA256: %s' % sha256)


class SubmissionException(Exception):
    pass


class SubmissionClient:
    """A helper class to simplify submitting files from internal or external sources.

    This tool helps take care of interactions between the filestore,
    datastore, dispatcher, and any sources of files to be processed.
    """

    def __init__(self, datastore: AssemblylineDatastore = None, filestore: FileStore = None,
                 config=None, redis=None, identify=None):
        self.log = logging.getLogger('assemblyline.submission_client')
        self.config = config or forge.CachedObject(forge.get_config)
        self.datastore = datastore or forge.get_datastore(self.config)
        self.filestore = filestore or forge.get_filestore(self.config)
        self.redis = redis
        if identify:
            self.cleanup = False
        else:
            self.cleanup = True
        self.identify = identify or forge.get_identify(config=self.config, datastore=self.datastore, use_cache=True)

        # A client for interacting with the dispatcher
        self.dispatcher = DispatchClient(datastore, redis)

    def __enter__(self):
        return self

    def __exit__(self, *_):
        self.stop()

    def stop(self):
        if self.cleanup:
            self.identify.stop()

    @elasticapm.capture_span(span_type='submission_client')
    def rescan(self, submission: Submission, results: Dict[str, Result], file_infos: Dict[str, FileInfo],
               file_tree, errors: List[str],  rescan_services: List[str]):
        """
        Rescan a submission started on another system.
        """
        # Reset submission processing data
        submission['times'].pop('completed')
        submission['state'] = 'submitted'

        # Set the list of service to rescan
        submission['params']['services']['rescan'] = rescan_services

        # Create the submission object
        submission_obj = Submission(submission)

        if len(submission_obj.files) == 0:
            raise SubmissionException("No files found to submit.")

        for f in submission_obj.files:
            if not self.datastore.file.exists(f.sha256):
                raise SubmissionException(f"File {f.sha256} does not exist, cannot continue submission.")

        # Set the new expiry
        if submission_obj.params.ttl:
            submission_obj.expiry_ts = epoch_to_iso(now() + submission_obj.params.ttl * 24 * 60 * 60)

        # Clearing runtime_excluded on initial submit or resubmit
        submission_obj.params.services.runtime_excluded = []

        # Save the submission
        self.datastore.submission.save(submission_obj.sid, submission_obj)

        # Dispatch the submission
        self.log.debug("Submission complete. Dispatching: %s", submission_obj.sid)
        self.dispatcher.dispatch_bundle(submission_obj, results, file_infos, file_tree, errors)

        return submission

    @elasticapm.capture_span(span_type='submission_client')
    def submit(self, submission_obj: SubmissionObject, local_files: List = None, completed_queue=None):
        """Submit several files in a single submission.

        After this method runs, there should be no local copies of the file left.
        """
        if local_files is None:
            local_files = []

        if len(submission_obj.files) == 0 and len(local_files) == 0:
            raise SubmissionException("No files found to submit...")

        if submission_obj.params.ttl:
            expiry = epoch_to_iso(submission_obj.time.timestamp() + submission_obj.params.ttl * 24 * 60 * 60)
        else:
            expiry = None
        max_size = self.config.submission.max_file_size

        for local_file in local_files:
            if isinstance(local_file, tuple):
                fname, local_file = local_file
            else:
                fname = safe_str(os.path.basename(local_file))

            # Upload/download, extract, analyze files
            original_classification = str(submission_obj.params.classification)
            file_hash, size, new_metadata = self._ready_file(local_file, expiry, original_classification)
            new_name = new_metadata.pop('name', fname)
            meta_classification = new_metadata.pop('classification', original_classification)
            if meta_classification != original_classification:
                try:
                    submission_obj.params.classification = Classification.max_classification(
                        meta_classification, original_classification)
                except InvalidClassification as ic:
                    raise SubmissionException("The classification found inside the cart file cannot be merged with "
                                              f"the classification the file was submitted as: {str(ic)}")

            submission_obj.metadata.update(**flatten(new_metadata))

            # Check that after we have resolved exactly what to pass on, that it
            # remains a valid target for scanning
            if size > max_size and not submission_obj.params.ignore_size:
                msg = "File too large (%d > %d). Submission failed" % (size, max_size)
                raise SubmissionException(msg)
            elif size == 0:
                msg = "File empty. Submission failed"
                raise SubmissionException(msg)

            submission_obj.files.append(File({
                'name': new_name,
                'size': size,
                'sha256': file_hash,
            }))

        # Clearing runtime_excluded on initial submit or resubmit
        submission_obj.params.services.runtime_excluded = []

        # We should now have all the information we need to construct a submission object
        sub = Submission(dict(
            archive_ts=now_as_iso(self.config.datastore.ilm.days_until_archive * 24 * 60 * 60),
            classification=submission_obj.params.classification,
            error_count=0,
            errors=[],
            expiry_ts=expiry,
            file_count=len(submission_obj.files),
            files=submission_obj.files,
            max_score=0,
            metadata=submission_obj.metadata,
            params=submission_obj.params,
            results=[],
            sid=submission_obj.sid,
            state='submitted',
            scan_key=submission_obj.scan_key,
        ))

        if self.config.ui.allow_malicious_hinting and submission_obj.params.malicious:
            sub.verdict = {"malicious": [submission_obj.params.submitter]}

        self.datastore.submission.save(sub.sid, sub)

        self.log.debug("Submission complete. Dispatching: %s", sub.sid)
        self.dispatcher.dispatch_submission(sub, completed_queue=completed_queue)

        return sub

    def _ready_file(self, local_path: str, expiry, classification) -> Tuple[str, int, dict]:
        """Take a file from local storage and prepare it for submission.

        After this method finished the file will ONLY exist on the filestore, not locally.
        """
        extracted_path = None
        try:
            # Analyze the file and make sure the file table is up to date
            fileinfo = self.identify.fileinfo(local_path)

            if fileinfo['size'] == 0:
                raise SubmissionException("File empty. Submission failed")

            # Check if there is an integrated decode process for this file
            # eg. files that are packaged, and the contained file (not the package
            # that local_path points to) should be passed into the system.
            extracted_path, fileinfo, al_meta = decode_file(local_path, fileinfo, self.identify)
            al_meta['classification'] = al_meta.get('classification', classification)
            if not Classification.is_valid(al_meta['classification']):
                raise SubmissionException(f"{al_meta['classification']} is not a valid classification for this system"
                                          ", submission is cancelled...")

            if extracted_path:
                local_path = extracted_path

            self.datastore.save_or_freshen_file(fileinfo['sha256'], fileinfo, expiry,
                                                al_meta['classification'], redis=self.redis)
            self.filestore.upload(local_path, fileinfo['sha256'])
            return fileinfo['sha256'], fileinfo['size'], al_meta

        finally:
            # If we extracted anything delete it
            if extracted_path:
                if os.path.exists(extracted_path):
                    os.unlink(extracted_path)



import logging
import time
from unittest import mock

import json
import pytest

from assemblyline.common.forge import get_service_queue, get_classification
from assemblyline.odm.models.error import Error
from assemblyline.odm.models.file import File
from assemblyline.odm.models.result import Result
from assemblyline.odm.randomizer import random_model_obj, random_minimal_obj, get_random_hash
from assemblyline.odm import models
from assemblyline.common.metrics import MetricsFactory

from assemblyline_core.dispatching.client import DispatchClient
from assemblyline_core.dispatching.dispatcher import Dispatcher, Submission
from assemblyline_core.dispatching.schedules import Scheduler as RealScheduler

# noinspection PyUnresolvedReferences
from assemblyline_core.dispatching.timeout import TimeoutTable
from mocking import ToggleTrue
from test_scheduler import dummy_service


logger = logging.getLogger('assemblyline.test')


class Scheduler(RealScheduler):
    def __init__(self, *args, **kwargs):
        pass

    def build_schedule(self, *args):
        return [
            {'extract': '', 'wrench': ''},
            {'av-a': '', 'av-b': '', 'frankenstrings': ''},
            {'xerox': ''}
        ]

    @property
    def services(self):
        return {
            'extract': dummy_service('extract', 'pre', extra_data=True),
            'wrench': dummy_service('wrench', 'pre'),
            'av-a': dummy_service('av-a', 'core'),
            'av-b': dummy_service('av-b', 'core'),
            'frankenstrings': dummy_service('frankenstrings', 'core'),
            'xerox': dummy_service('xerox', 'post'),
        }


def make_result(file_hash, service):
    new_result: Result = random_minimal_obj(Result)
    new_result.sha256 = file_hash
    new_result.response.service_name = service
    return new_result


def make_error(file_hash, service, recoverable=True):
    new_error: Error = random_model_obj(Error)
    new_error.response.service_name = service
    new_error.sha256 = file_hash
    if recoverable:
        new_error.response.status = 'FAIL_RECOVERABLE'
    else:
        new_error.response.status = 'FAIL_NONRECOVERABLE'
    return new_error


def wait_result(task, file_hash, service):
    for _ in range(10):
        if (file_hash, service) in task.service_results:
            return True
        time.sleep(0.05)


def wait_error(task, file_hash, service):
    for _ in range(10):
        if (file_hash, service) in task.service_errors:
            return True
        time.sleep(0.05)


@pytest.fixture(autouse=True)
def log_config(caplog):
    caplog.set_level(logging.INFO, logger='assemblyline')
    from assemblyline.common import log as al_log
    al_log.init_logging = lambda *args: None


@mock.patch('assemblyline_core.dispatching.dispatcher.Scheduler', Scheduler)
@mock.patch('assemblyline_core.dispatching.dispatcher.MetricsFactory', new=mock.MagicMock(spec=MetricsFactory))
def test_simple(clean_redis, clean_datastore):
    ds = clean_datastore
    redis = clean_redis

    def service_queue(name): return get_service_queue(name, redis)

    file = random_model_obj(File)
    file_hash = file.sha256
    file.type = 'unknown'
    ds.file.save(file_hash, file)

    sub: Submission = random_model_obj(models.submission.Submission)
    sub.sid = sid = 'first-submission'
    sub.params.ignore_cache = False
    sub.params.max_extracted = 5
    sub.params.classification = get_classification().UNRESTRICTED
    sub.params.initial_data = json.dumps({'cats': 'big'})
    sub.files = [dict(sha256=file_hash, name='file')]

    disp = Dispatcher(ds, redis, redis)
    disp.running = ToggleTrue()
    client = DispatchClient(ds, redis, redis)
    client.dispatcher_data_age = time.time()
    client.dispatcher_data.append(disp.instance_id)

    # Submit a problem, and check that it gets added to the dispatch hash
    # and the right service queues
    logger.info('==== first dispatch')
    # task = SubmissionTask(sub.as_primitives(), 'some-completion-queue')
    client.dispatch_submission(sub)
    disp.pull_submissions()
    disp.service_worker(disp.process_queue_index(sid))
    task = disp.tasks.get(sid)

    assert task.queue_keys[(file_hash, 'extract')] is not None
    assert task.queue_keys[(file_hash, 'wrench')] is not None
    assert service_queue('extract').length() == 1
    assert service_queue('wrench').length() == 1

    # Making the same call again will queue it up again
    logger.info('==== second dispatch')
    disp.dispatch_file(task, file_hash)

    assert task.queue_keys[(file_hash, 'extract')] is not None
    assert task.queue_keys[(file_hash, 'wrench')] is not None
    assert service_queue('extract').length() == 1  # the queue doesn't pile up
    assert service_queue('wrench').length() == 1

    logger.info('==== third dispatch')
    job = client.request_work('0', 'extract', '0')
    assert job.temporary_submission_data == [{'name': 'cats', 'value': 'big'}]
    client.service_failed(sid, 'abc123', make_error(file_hash, 'extract'))
    # Deliberately do in the wrong order to make sure that works
    disp.pull_service_results()
    disp.service_worker(disp.process_queue_index(sid))

    assert task.queue_keys[(file_hash, 'extract')] is not None
    assert task.queue_keys[(file_hash, 'wrench')] is not None
    assert service_queue('extract').length() == 1

    # Mark extract as finished, wrench as failed
    logger.info('==== fourth dispatch')
    client.request_work('0', 'extract', '0')
    client.request_work('0', 'wrench', '0')
    client.service_finished(sid, 'extract-result', make_result(file_hash, 'extract'))
    client.service_failed(sid, 'wrench-error', make_error(file_hash, 'wrench', False))
    for _ in range(2):
        disp.pull_service_results()
        disp.service_worker(disp.process_queue_index(sid))

    assert wait_error(task, file_hash, 'wrench')
    assert wait_result(task, file_hash, 'extract')
    assert service_queue('av-a').length() == 1
    assert service_queue('av-b').length() == 1
    assert service_queue('frankenstrings').length() == 1

    # Have the AVs fail, frankenstrings finishes
    logger.info('==== fifth dispatch')
    client.request_work('0', 'av-a', '0')
    client.request_work('0', 'av-b', '0')
    client.request_work('0', 'frankenstrings', '0')
    client.service_failed(sid, 'av-a-error', make_error(file_hash, 'av-a', False))
    client.service_failed(sid, 'av-b-error', make_error(file_hash, 'av-b', False))
    client.service_finished(sid, 'f-result', make_result(file_hash, 'frankenstrings'))
    for _ in range(3):
        disp.pull_service_results()
        disp.service_worker(disp.process_queue_index(sid))

    assert wait_result(task, file_hash, 'frankenstrings')
    assert wait_error(task, file_hash, 'av-a')
    assert wait_error(task, file_hash, 'av-b')
    assert service_queue('xerox').length() == 1

    # Finish the xerox service and check if the submission completion got checked
    logger.info('==== sixth dispatch')
    client.request_work('0', 'xerox', '0')
    client.service_finished(sid, 'xerox-result-key', make_result(file_hash, 'xerox'))
    disp.pull_service_results()
    disp.service_worker(disp.process_queue_index(sid))
    disp.save_submission()

    assert wait_result(task, file_hash, 'xerox')
    assert disp.tasks.get(sid) is None


@mock.patch('assemblyline_core.dispatching.dispatcher.MetricsFactory', mock.MagicMock())
@mock.patch('assemblyline_core.dispatching.dispatcher.Scheduler', Scheduler)
def test_dispatch_extracted(clean_redis, clean_datastore):
    redis = clean_redis
    ds = clean_datastore

    # def service_queue(name): return get_service_queue(name, redis)

    # Setup the fake datastore
    file_hash = get_random_hash(64)
    second_file_hash = get_random_hash(64)

    for fh in [file_hash, second_file_hash]:
        obj = random_model_obj(models.file.File)
        obj.sha256 = fh
        ds.file.save(fh, obj)

    # Inject the fake submission
    submission = random_model_obj(models.submission.Submission)
    submission.files = [dict(name='./file', sha256=file_hash)]
    sid = submission.sid = 'first-submission'

    disp = Dispatcher(ds, redis, redis)
    disp.running = ToggleTrue()
    client = DispatchClient(ds, redis, redis)
    client.dispatcher_data_age = time.time()
    client.dispatcher_data.append(disp.instance_id)

    # Launch the submission
    client.dispatch_submission(submission)
    disp.pull_submissions()
    disp.service_worker(disp.process_queue_index(sid))

    # Finish one service extracting a file
    job = client.request_work('0', 'extract', '0')
    assert job.fileinfo.sha256 == file_hash
    assert job.filename == './file'
    new_result: Result = random_minimal_obj(Result)
    new_result.sha256 = file_hash
    new_result.response.service_name = 'extract'
    new_result.response.extracted = [dict(sha256=second_file_hash, name='second-*',
                                          description='abc', classification='U')]
    client.service_finished(sid, 'extracted-done', new_result)

    # process the result
    disp.pull_service_results()
    disp.service_worker(disp.process_queue_index(sid))
    disp.service_worker(disp.process_queue_index(sid))

    #
    job = client.request_work('0', 'extract', '0')
    assert job.fileinfo.sha256 == second_file_hash
    assert job.filename == 'second-*'


mock_time = mock.Mock()
mock_time.return_value = 0


@mock.patch('time.time', mock_time)
def test_timeout():
    table = TimeoutTable()
    table.set('first', 5, 1)
    table.set('second', 10, 2)
    table.set('third', 15, 3)
    table.set('fourth', 20, 4)

    # Expire one thing
    mock_time.return_value = 6
    items = table.timeouts()
    assert len(items) == 1
    assert items['first'] == 1

    # Replace the data and expiry
    table.set('second', 20, 5)  # second now expires at 26

    # Expire two things
    mock_time.return_value = 21
    items = table.timeouts()
    assert len(items) == 2
    assert items['third'] == 3
    assert items['fourth'] == 4

    # Expire nothing
    assert len(table.timeouts()) == 0

    # Expire the final thing
    mock_time.return_value = 30
    items = table.timeouts()
    assert len(items) == 1
    assert items['second'] == 5

    # Expire nothing
    assert len(table.timeouts()) == 0



#!/usr/bin/env python
from __future__ import annotations
import concurrent.futures
from concurrent.futures import ThreadPoolExecutor, ProcessPoolExecutor, Future, as_completed
from concurrent.futures.process import BrokenProcessPool
import functools
from typing import Callable, Optional, Union, TYPE_CHECKING
import elasticapm
import time

from datemath import dm

from assemblyline.common.isotime import epoch_to_iso, now_as_iso
from assemblyline.datastore.helper import AssemblylineDatastore
from assemblyline.datastore.store import ESStore
from assemblyline_core.server_base import ServerBase
from assemblyline.common import forge
from assemblyline.common.metrics import MetricsFactory
from assemblyline.filestore import FileStore
from assemblyline.odm.messages.expiry_heartbeat import Metrics

if TYPE_CHECKING:
    from assemblyline.datastore.collection import ESCollection


DAY_SECONDS = 24 * 60 * 60


def file_archive_delete_worker(logger, datastore_url, collection_name,
                               filestore_urls, file_batch, expiry_time) -> list[str]:

    datastore = AssemblylineDatastore(ESStore(datastore_url, archive_access=True))
    collection = datastore.get_collection(collection_name)

    try:
        filestore = FileStore(*filestore_urls)

        def archive_filestore_delete(sha256: str) -> Optional[str]:
            # If we are working with an archive, their may be a hot entry that expires later.
            doc = collection.get_if_exists(sha256, as_obj=False)
            if doc and doc['expiry_ts'] > expiry_time:
                return None
            filestore.delete(sha256)
            if not filestore.exists(sha256):
                return sha256
            return None

        return _file_delete_worker(logger, archive_filestore_delete, file_batch)

    except Exception as error:
        logger.exception("Error in filestore worker: " + str(error))
    return []


def file_delete_worker(logger, filestore_urls, file_batch) -> list[str]:
    try:
        filestore = FileStore(*filestore_urls)

        def filestore_delete(sha256: str) -> Optional[str]:
            filestore.delete(sha256)
            if not filestore.exists(sha256):
                return sha256
            return None

        return _file_delete_worker(logger, filestore_delete, file_batch)

    except Exception as error:
        logger.exception("Error in filestore worker: " + str(error))
    return []


def _file_delete_worker(logger, delete_action: Callable[[str], Optional[str]], file_batch) -> list[str]:
    finished_files: list[str] = []
    try:
        futures = []

        with ThreadPoolExecutor(8) as pool:
            for filename in file_batch:
                futures.append(pool.submit(delete_action, filename))

            for future in as_completed(futures):
                try:
                    erased_name = future.result()
                    if erased_name:
                        finished_files.append(erased_name)
                except Exception as error:
                    logger.exception("Error in filestore worker: " + str(error))

    except Exception as error:
        logger.exception("Error in filestore worker: " + str(error))
    return finished_files


class ExpiryManager(ServerBase):
    def __init__(self, force_ilm=False):
        self.config = forge.get_config()
        if force_ilm:
            self.config.datastore.ilm.enabled = True

        super().__init__('assemblyline.expiry', shutdown_timeout=self.config.core.expiry.sleep_time + 5)
        self.datastore = forge.get_datastore(config=self.config, archive_access=True)
        self.hot_datastore = forge.get_datastore(config=self.config, archive_access=False)
        self.filestore = forge.get_filestore(config=self.config)
        self.cachestore = FileStore(*self.config.filestore.cache)
        self.expirable_collections: list[ESCollection] = []
        self.archiveable_collections: list[ESCollection] = []
        self.counter = MetricsFactory('expiry', Metrics)
        self.counter_archive = MetricsFactory('archive', Metrics)
        self.file_delete_worker = ProcessPoolExecutor(self.config.core.expiry.delete_workers)

        if self.config.datastore.ilm.enabled:
            self.fs_hashmap = {
                'file': self.archive_filestore_delete,
                'cached_file': self.archive_cachestore_delete
            }
        else:
            self.fs_hashmap = {
                'file': self.filestore_delete,
                'cached_file': self.cachestore_delete
            }

        for name, definition in self.datastore.ds.get_models().items():
            if hasattr(definition, 'archive_ts'):
                self.archiveable_collections.append(getattr(self.datastore, name))
            if hasattr(definition, 'expiry_ts'):
                self.expirable_collections.append(getattr(self.datastore, name))

        if self.config.core.metrics.apm_server.server_url is not None:
            self.log.info(f"Exporting application metrics to: {self.config.core.metrics.apm_server.server_url}")
            elasticapm.instrument()
            self.apm_client = elasticapm.Client(server_url=self.config.core.metrics.apm_server.server_url,
                                                service_name="expiry")
        else:
            self.apm_client = None

    @property
    def expiry_size(self):
        return self.config.core.expiry.delete_batch_size

    @property
    def archive_size(self):
        return self.config.core.expiry.archive_batch_size

    def stop(self):
        if self.counter:
            self.counter.stop()

        if self.apm_client:
            elasticapm.uninstrument()
        super().stop()

    def log_errors(self, function):
        @functools.wraps(function)
        def _func(*args, **kwargs):
            try:
                function(*args, **kwargs)
            except Exception:
                self.log.exception("Error in expiry worker")
        return _func

    def filestore_delete(self, file_batch, _):
        return self.file_delete_worker.submit(file_delete_worker, logger=self.log,
                                              filestore_urls=list(self.config.filestore.storage),
                                              file_batch=file_batch)

    def archive_filestore_delete(self, file_batch, expiry_time):
        return self.file_delete_worker.submit(file_archive_delete_worker, collection_name='file',
                                              logger=self.log, datastore_url=list(self.config.datastore.hosts),
                                              filestore_urls=list(self.config.filestore.storage),
                                              file_batch=file_batch, expiry_time=expiry_time)

    def cachestore_delete(self, file_batch, _):
        return self.file_delete_worker.submit(file_delete_worker, logger=self.log,
                                              filestore_urls=list(self.config.filestore.cache),
                                              file_batch=file_batch)

    def archive_cachestore_delete(self, file_batch, expiry_time):
        return self.file_delete_worker.submit(file_archive_delete_worker, collection_name='cached_file',
                                              logger=self.log, datastore_url=list(self.config.datastore.hosts),
                                              filestore_urls=list(self.config.filestore.cache),
                                              file_batch=file_batch, expiry_time=expiry_time)

    def _finish_delete(self, collection: ESCollection, task: Future):
        # Wait until the worker process finishes deleting files
        file_list: list[str] = []
        while self.running:
            self.heartbeat()
            try:
                file_list = task.result(5)
                break
            except concurrent.futures.TimeoutError:
                pass

        # build a batch delete job for all the removed files
        bulk = collection.get_bulk_plan()
        for sha256 in file_list:
            bulk.add_delete_operation(sha256)

        if len(file_list) > 0:
            self.log.info(f'    Deleted associated files from the '
                          f'{"cachestore" if "cache" in collection.name else "filestore"}...')
            collection.bulk(bulk)
            self.counter.increment(f'{collection.name}', increment_by=len(file_list))
            self.log.info(f"    Deleted {len(file_list)} items from the datastore...")
        else:
            self.log.warning('    Expiry unable to clean up any of the files in filestore.')

    def _simple_delete(self, collection, delete_query, number_to_delete):
        self.heartbeat()
        collection.delete_by_query(delete_query, workers=self.config.core.expiry.workers)
        self.counter.increment(f'{collection.name}', increment_by=number_to_delete)
        self.log.info(f"    Deleted {number_to_delete} items from the datastore...")

    def run_expiry_once(self, pool: ThreadPoolExecutor):
        now = now_as_iso()
        reached_max = False

        # Expire data
        for collection in self.expirable_collections:
            # Call heartbeat pre-dated by 5 minutes. If a collection takes more than
            # 5 minutes to expire, this container could be seen as unhealthy. The down
            # side is if it is stuck on something it will be more than 5 minutes before
            # the container is restarted.
            self.heartbeat(int(time.time() + 5 * 60))

            # Start of expiry transaction
            if self.apm_client:
                self.apm_client.begin_transaction("Delete expired documents")

            if self.config.core.expiry.batch_delete:
                final_date = dm(f"{now}||-{self.config.core.expiry.delay}h/d").float_timestamp
            else:
                final_date = dm(f"{now}||-{self.config.core.expiry.delay}h").float_timestamp
            final_date_string = epoch_to_iso(final_date)

            # Break down the expiry window into smaller chunks of data
            unchecked_chunks: list[tuple[float, float]] = [(0, final_date)]
            ready_chunks: dict[tuple[float, float], int] = {}
            while unchecked_chunks and len(ready_chunks) < self.config.core.expiry.iteration_max_tasks:
                start, end = unchecked_chunks.pop(0)
                chunk_size = self._count_expired(collection, start, end)

                # Empty chunks are fine
                if chunk_size == 0:
                    continue

                # We found a small enough chunk to
                #  run on
                if chunk_size < self.expiry_size:
                    ready_chunks[(start, end)] = chunk_size
                    continue

                # Break this chunk into parts
                if start == 0:
                    middle = end - DAY_SECONDS
                else:
                    middle = (end + start)/2
                unchecked_chunks.append((middle, end))
                unchecked_chunks.append((start, middle))

            # If there are still chunks we haven't checked, then we know there is more data
            if unchecked_chunks:
                reached_max = True

            for (start, end), number_to_delete in ready_chunks.items():
                # We assume that no records are ever inserted such that their expiry_ts is in the past.
                # We also assume that the `end` dates are also in the past.
                # As long as these two things are true, the set returned by this query should be consistent.
                # The one race condition is that a record might be refreshed while the file
                # blob would be deleted anyway, leaving a file record with no filestore object
                self.log.info(f"Processing collection: {collection.name}")
                delete_query = f"expiry_ts:[{epoch_to_iso(start) if start > 0 else '*'} TO {epoch_to_iso(end)}}}"

                # check if we are dealing with an index that needs file cleanup
                if self.config.core.expiry.delete_storage and collection.name in self.fs_hashmap:
                    # Delete associated files
                    delete_objects: list[str] = []
                    for item in collection.stream_search(delete_query, fl='id', use_archive=True, as_obj=False):
                        delete_objects.append(item['id'])
                    delete_task = self.fs_hashmap[collection.name](delete_objects, final_date_string)

                    # Proceed with deletion, but only after all the scheduled deletes for this
                    self.log.info(f"Scheduled {len(delete_objects)}/{number_to_delete} "
                                  f"files to be removed for: {collection.name}")
                    pool.submit(self.log_errors(self._finish_delete), collection, delete_task)

                else:
                    # Proceed with deletion
                    pool.submit(self.log_errors(self._simple_delete),
                                collection, delete_query, number_to_delete)

            # End of expiry transaction
            if self.apm_client:
                self.apm_client.end_transaction(collection.name, 'deleted')

        return reached_max

    def run_archive_once(self, pool: ThreadPoolExecutor):
        reached_max = False
        if not self.config.datastore.ilm.enabled:
            return reached_max

        # Start of expiry transaction
        if self.apm_client:
            self.apm_client.begin_transaction("Archive older documents")

        # Collect a set of jobs which archive all the collections in
        # managable sized queries. Breaking up the query is helpful
        # where each of our archive calls are actually a reindex
        # followed by a delete, which we would like to both complete
        # relatively close togeather.
        for collection in self.archiveable_collections:
            self.heartbeat(int(time.time() + 60))
            reached_max |= self._archive_collection(collection, pool)

        # End of expiry transaction
        if self.apm_client:
            self.apm_client.end_transaction(result='archived')

        return reached_max

    def _find_archive_start(self, container: ESCollection):
        """
        Moving backwards one day at a time get a rough idea where archiveable material
        in the datastore starts.
        """
        now = time.time()
        offset = 1
        while True:
            count = self._count_archivable(container, "*", now - offset * DAY_SECONDS)
            if count == 0:
                return now - offset * DAY_SECONDS
            offset += 1

    def _count_expired(self, container: ESCollection, start: Union[float, str], end: float) -> int:
        """Count how many items need to be erased in the given window."""
        if start == 0:
            start = '*'
        if isinstance(start, (float, int)):
            start = epoch_to_iso(start)
        query = f'expiry_ts:[{start} TO {epoch_to_iso(end)}}}'

        return container.search(query, rows=0, as_obj=False, use_archive=True, track_total_hits=self.expiry_size)['total']

    def _count_archivable(self, container: ESCollection, start: Union[float, str], end: float) -> int:
        """Count how many items need to be archived in the given window."""
        if isinstance(start, (float, int)):
            start = epoch_to_iso(start)
        query = f'archive_ts:[{start} TO {epoch_to_iso(end)}}}'
        return container.search(query, rows=0, as_obj=False, use_archive=False, track_total_hits=self.archive_size)['total']

    def _archive_collection(self, collection: ESCollection, pool: ThreadPoolExecutor) -> bool:
        # Start with archiving everything up until now
        chunks: list[tuple[float, float]] = [(self._find_archive_start(collection), time.time())]
        futures: list[Future] = []

        while len(chunks) > 0 and len(futures) < self.config.core.expiry.iteration_max_tasks:
            # Take the next chunk, and figure out how many records it covers
            start, end = chunks.pop(0)
            count = self._count_archivable(collection, start, end)

            # Chunks that are fully archived are fine to skip
            if count == 0:
                continue

            # If the chunk is bigger than the number we intend to archive
            # in a single call break it into parts
            if count >= self.archive_size:
                middle = (start + end)/2
                chunks.append((start, middle))
                chunks.append((middle, end))
                continue

            # Schedule the chunk of data to be archived
            @self.log_errors
            def archive_chunk(query, expected):
                self.heartbeat()
                if collection.archive(query):
                    self.counter_archive.increment(f'{collection.name}', increment_by=expected)
                else:
                    self.log.error(f"Failed to archive documents matching: '{query}'")

            query = f'archive_ts:[{epoch_to_iso(start)} TO {epoch_to_iso(end)}}}'
            futures.append(pool.submit(archive_chunk, query, count))

        return len(futures) == self.config.core.expiry.iteration_max_tasks

    def try_run(self):
        while self.running:
            try:
                expiry_maxed_out = False
                archive_maxed_out = False

                with ThreadPoolExecutor(self.config.core.expiry.workers) as pool:
                    try:
                        expiry_maxed_out = self.run_expiry_once(pool)
                    except Exception as e:
                        self.log.exception(str(e))

                    try:
                        archive_maxed_out = self.run_archive_once(pool)
                    except Exception as e:
                        self.log.exception(str(e))

                if not expiry_maxed_out and not archive_maxed_out:
                    self.sleep_with_heartbeat(self.config.core.expiry.sleep_time)

            except BrokenProcessPool:
                self.log.error("File delete worker pool crashed.")
                self.file_delete_worker = ProcessPoolExecutor(self.config.core.expiry.delete_workers)


if __name__ == "__main__":
    with ExpiryManager() as em:
        em.serve_forever()



import os
import time

from assemblyline.common import forge
from assemblyline.common.bundling import create_bundle, import_bundle
from assemblyline.remote.datatypes import get_client
from assemblyline.remote.datatypes.queues.named import NamedQueue


EMPTY_WAIT_TIME = int(os.environ.get('EMPTY_WAIT_TIME', '30'))
REPLAY_REQUESTED = 'requested'
REPLAY_PENDING = 'pending'
REPLAY_DONE = 'done'


class ClientBase(object):
    def __init__(self, log, alert_fqs=None, submission_fqs=None, lookback_time='*'):
        # Set logger
        self.log = log

        # Setup timming
        self.last_alert_time = self.last_submission_time = self.lookback_time = lookback_time

        # Setup filter queries
        self.pending_fq = f'NOT metadata.replay:{REPLAY_PENDING}'
        self.done_fq = f'NOT metadata.replay:{REPLAY_DONE}'
        self.alert_fqs = alert_fqs or []
        self.submission_fqs = submission_fqs or []

        # Set running flag
        self.running = True

    def _get_next_alert_ids(self, *_):
        raise NotImplementedError()

    def _get_next_submission_ids(self, *_):
        raise NotImplementedError()

    def _set_bulk_alert_pending(self, *_):
        raise NotImplementedError()

    def _set_bulk_submission_pending(self, *_):
        raise NotImplementedError()

    def _stream_alert_ids(self, *_):
        raise NotImplementedError()

    def _stream_submission_ids(self, *_):
        raise NotImplementedError()

    def create_alert_bundle(self, *_):
        raise NotImplementedError()

    def create_submission_bundle(self, *_):
        raise NotImplementedError()

    def load_bundle(self, *_):
        raise NotImplementedError()

    def stop(self):
        self.running = False

    def set_single_alert_complete(self, *_):
        raise NotImplementedError()

    def set_single_submission_complete(self, *_):
        raise NotImplementedError()

    def setup_alert_input_queue(self, once=False):
        # Bootstrap recovery of pending replayed alerts
        for a in self._stream_alert_ids(f"metadata.replay:{REPLAY_PENDING}"):
            self.log.info(f"Replaying alert: {a['alert_id']}")
            self.put_alert(a)

        # Create the list of filter queries
        processing_fqs = self.alert_fqs + [self.pending_fq, self.done_fq]

        # Run
        while self.running:
            # Find alerts
            alert_input_query = f"reporting_ts:{{{self.last_alert_time} TO now]"
            alerts = self._get_next_alert_ids(alert_input_query, processing_fqs)

            # Set their pending state
            if alerts['items']:
                last_time = alerts['items'][-1]['reporting_ts']
                bulk_query = f"reporting_ts:{{{self.last_alert_time} TO {last_time}]"
                count = len(alerts['items'])
                self._set_bulk_alert_pending(bulk_query, processing_fqs, count)
                self.last_alert_time = last_time

            # Queue them
            for a in alerts['items']:
                self.log.info(f"Replaying alert: {a['alert_id']}")
                self.put_alert(a)

            # Wait if nothing found
            if alerts['total'] == 0:
                self.last_alert_time = self.lookback_time
                for _ in range(EMPTY_WAIT_TIME):
                    if not self.running:
                        break
                    time.sleep(1)

            if once:
                break

    def setup_submission_input_queue(self, once=False):
        # Bootstrap recovery of pending replayed submission
        for sub in self._stream_submission_ids(f"metadata.replay:{REPLAY_PENDING}"):
            self.log.info(f"Replaying submission: {sub['sid']}")
            self.put_submission(sub)

        # Create the list of filter queries
        processing_fqs = self.submission_fqs + [self.pending_fq, self.done_fq]

        # Run
        while self.running:
            # Find submissions
            sub_query = f"times.completed:{{{self.last_submission_time} TO now]"
            submissions = self._get_next_submission_ids(sub_query, processing_fqs)

            # Set their pending state
            if submissions['items']:
                last_time = submissions['items'][-1]['times']['completed']
                bulk_query = f"times.completed:{{{self.last_submission_time} TO {last_time}]"
                count = len(submissions['items'])
                self._set_bulk_submission_pending(bulk_query, processing_fqs, count)
                self.last_submission_time = last_time

            # Queue them
            for sub in submissions['items']:
                self.log.info(f"Replaying submission: {sub['sid']}")
                self.put_submission(sub)

            # Wait if nothing found
            if submissions['total'] == 0:
                self.last_submission_time = self.lookback_time
                for _ in range(EMPTY_WAIT_TIME):
                    if not self.running:
                        break
                    time.sleep(1)

            if once:
                break

    def get_next_alert(self):
        raise NotImplementedError()

    def get_next_submission(self):
        raise NotImplementedError()

    def put_alert(self, *_):
        raise NotImplementedError()

    def put_submission(self, *_):
        raise NotImplementedError()

    def get_next_file(self):
        raise NotImplementedError()

    def put_file(self, *_):
        raise NotImplementedError()


class APIClient(ClientBase):
    def __init__(self, log, host, user, apikey, verify, alert_fqs=None, submission_fqs=None, lookback_time='*'):
        # Import assemblyline client
        from assemblyline_client import get_client

        # Setup AL client
        self.al_client = get_client(host, apikey=(user, apikey), verify=verify)

        super().__init__(log, alert_fqs=alert_fqs, submission_fqs=submission_fqs, lookback_time=lookback_time)

    def _get_next_alert_ids(self, query, filter_queries):
        return self.al_client.search.alert(
            query, fl="alert_id,reporting_ts", sort="reporting_ts asc", rows=100, filters=filter_queries)

    def _get_next_submission_ids(self, query, filter_queries):
        return self.al_client.search.submission(
            query, fl="sid,times.completed", sort="times.completed asc", rows=100, filters=filter_queries)

    def _set_bulk_alert_pending(self, query, filter_queries, max_docs):
        self.al_client.replay.set_bulk_pending('alert', query, filter_queries, max_docs)

    def _set_bulk_submission_pending(self, query, filter_queries, max_docs):
        self.al_client.replay.set_bulk_pending('submission', query, filter_queries, max_docs)

    def _stream_alert_ids(self, query):
        return self.al_client.search.stream.alert(query, fl="alert_id,reporting_ts")

    def _stream_submission_ids(self, query):
        return self.al_client.search.stream.submission(query, fl="sid,times.completed")

    def create_alert_bundle(self, alert_id, bundle_path):
        self.al_client.bundle.create(alert_id, output=bundle_path, use_alert=True)

    def create_submission_bundle(self, sid, bundle_path):
        self.al_client.bundle.create(sid, output=bundle_path)

    def load_bundle(self, bundle_path, min_classification, rescan_services, exist_ok=True):
        self.al_client.bundle.import_bundle(bundle_path,
                                            min_classification=min_classification,
                                            rescan_services=rescan_services,
                                            exist_ok=exist_ok)

    def set_single_alert_complete(self, alert_id):
        self.al_client.replay.set_complete('alert', alert_id)

    def set_single_submission_complete(self, sid):
        self.al_client.replay.set_complete('submission', sid)

    def get_next_alert(self):
        return self.al_client.replay.get_message('alert')

    def get_next_file(self):
        return self.al_client.replay.get_message('file')

    def get_next_submission(self):
        return self.al_client.replay.get_message('submission')

    def put_alert(self, alert):
        self.al_client.replay.put_message('alert', alert)

    def put_file(self, path):
        self.al_client.replay.put_message('file', path)

    def put_submission(self, submission):
        self.al_client.replay.put_message('submission', submission)


class DirectClient(ClientBase):
    def __init__(self, log, alert_fqs=None, submission_fqs=None, lookback_time='*'):
        # Setup datastore
        config = forge.get_config()
        redis = get_client(config.core.redis.nonpersistent.host, config.core.redis.nonpersistent.port, False)

        self.datastore = forge.get_datastore(config=config)
        self.alert_queue = NamedQueue("replay_alert", host=redis)
        self.file_queue = NamedQueue("replay_file", host=redis)
        self.submission_queue = NamedQueue("replay_submission", host=redis)

        super().__init__(log, alert_fqs=alert_fqs, submission_fqs=submission_fqs, lookback_time=lookback_time)

    def _get_next_alert_ids(self, query, filter_queries):
        return self.datastore.alert.search(
            query, fl="alert_id,reporting_ts", sort="reporting_ts asc", rows=100, as_obj=False, filters=filter_queries)

    def _get_next_submission_ids(self, query, filter_queries):
        return self.datastore.submission.search(
            query, fl="sid,times.completed", sort="times.completed asc", rows=100, as_obj=False, filters=filter_queries)

    def _set_bulk_alert_pending(self, query, filter_queries, max_docs):
        operations = [(self.datastore.alert.UPDATE_SET, 'metadata.replay', REPLAY_PENDING)]
        self.datastore.alert.update_by_query(query, operations, filters=filter_queries, max_docs=max_docs)

    def _set_bulk_submission_pending(self, query, filter_queries, max_docs):
        operations = [(self.datastore.submission.UPDATE_SET, 'metadata.replay', REPLAY_PENDING)]
        self.datastore.submission.update_by_query(query, operations, filters=filter_queries, max_docs=max_docs)

    def _stream_alert_ids(self, query):
        return self.datastore.alert.stream_search(query, fl="alert_id,reporting_ts", as_obj=False)

    def _stream_submission_ids(self, query):
        return self.datastore.submission.stream_search(query, fl="sid,times.completed", as_obj=False)

    def create_alert_bundle(self, alert_id, bundle_path):
        temp_bundle_file = create_bundle(alert_id, working_dir=os.path.dirname(bundle_path), use_alert=True)
        os.rename(temp_bundle_file, bundle_path)

    def create_submission_bundle(self, sid, bundle_path):
        temp_bundle_file = create_bundle(sid, working_dir=os.path.dirname(bundle_path))
        os.rename(temp_bundle_file, bundle_path)

    def load_bundle(self, bundle_path, min_classification, rescan_services, exist_ok=True):
        import_bundle(bundle_path,
                      min_classification=min_classification,
                      rescan_services=rescan_services,
                      exist_ok=exist_ok)

    def set_single_alert_complete(self, alert_id):
        operations = [(self.datastore.alert.UPDATE_SET, 'metadata.replay', REPLAY_DONE)]
        self.datastore.alert.update(alert_id, operations)

    def set_single_submission_complete(self, sid):
        operations = [(self.datastore.submission.UPDATE_SET, 'metadata.replay', REPLAY_DONE)]
        self.datastore.submission.update(sid, operations)

    def get_next_alert(self):
        return self.alert_queue.pop(blocking=True, timeout=30)

    def get_next_file(self):
        return self.file_queue.pop(blocking=True, timeout=30)

    def get_next_submission(self):
        return self.submission_queue.pop(blocking=True, timeout=30)

    def put_alert(self, alert):
        self.alert_queue.push(alert)

    def put_file(self, path):
        self.file_queue.push(path)

    def put_submission(self, submission):
        self.submission_queue.push(submission)



import logging

import yaml

from assemblyline.common import forge
from assemblyline.datastore.helper import AssemblylineDatastore
from assemblyline.filestore import FileStore


# Tasking class
class SafelistClient:
    """A helper class to simplify safelisting for privileged services and service-server."""

    def __init__(self, datastore: AssemblylineDatastore = None, config=None):
        self.log = logging.getLogger('assemblyline.safelist_client')
        self.config = config or forge.CachedObject(forge.get_config)
        self.datastore = datastore or forge.get_datastore(self.config)

    # Safelist
    def exists(self, qhash):
        return self.datastore.safelist.get_if_exists(qhash, as_obj=False)

    def get_safelisted_tags(self, tag_types):
        if isinstance(tag_types, str):
            tag_types = tag_types.split(',')

        with forge.get_cachestore('system', config=self.config, datastore=self.datastore) as cache:
            tag_safelist_yml = cache.get('tag_safelist_yml')
            if tag_safelist_yml:
                tag_safelist_data = yaml.safe_load(tag_safelist_yml)
            else:
                tag_safelist_data = forge.get_tag_safelist_data()

        if tag_types:
            output = {
                'match': {k: v for k, v in tag_safelist_data.get('match', {}).items()
                          if k in tag_types or tag_types == []},
                'regex': {k: v for k, v in tag_safelist_data.get('regex', {}).items()
                          if k in tag_types or tag_types == []},
            }
            for tag in tag_types:
                for sl in self.datastore.safelist.stream_search(
                        f"type:tag AND enabled:true AND tag.type:{tag}", as_obj=False):
                    output['match'].setdefault(sl['tag']['type'], [])
                    output['match'][sl['tag']['type']].append(sl['tag']['value'])

        else:
            output = tag_safelist_data
            for sl in self.datastore.safelist.stream_search("type:tag AND enabled:true", as_obj=False):
                output['match'].setdefault(sl['tag']['type'], [])
                output['match'][sl['tag']['type']].append(sl['tag']['value'])

        return output

    def get_safelisted_signatures(self, **_):
        output = [
            item['signature']['name']
            for item in self.datastore.safelist.stream_search(
                "type:signature AND enabled:true", fl="signature.name", as_obj=False)]

        return output
